using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;
using System.Threading;
using System.Diagnostics;

namespace Bliss.Threading
{
	/// <summary>
	/// Creates a thread with ApartmentState.MTA for invoking a specified method.
	/// Used by STA Proxy classes.
	/// </summary>
	[Author("Anton Delsink", "anton@delsink.com")]
	internal static class MtaThreadUtil
	{
		/// <summary>
		/// Helper delegate for jumping from one thread to another.
		/// </summary>
		/// <typeparam name="ReturnType">The data type of the target method's result.</typeparam>
		internal delegate ReturnType NoArgReturningDelegate<ReturnType>();
		internal delegate ReturnType SingleArgReturningDelegate<ReturnType, ArgType>(ArgType arg);
		internal delegate ReturnType TwoArgReturningDelegate<ReturnType, ArgOneType, ArgTwoType>(ArgOneType a1, ArgTwoType a2);
		internal delegate ReturnType ThreeArgReturningDelegate<ReturnType, ArgOneType, ArgTwoType, ArgThreeType>(ArgOneType a1, ArgTwoType a2, ArgThreeType a3);
		//internal delegate void NoArgActionDelegate(); // not in use
		internal delegate void SingleArgActionDelegate<ArgOneType>(ArgOneType a1);

		#region InvokeOnMtaThread
		internal static void InvokeOnMtaThread(ThreadStart targetMethod)
		{
			// TODO: Deal with exceptions

			Thread t = new Thread(new ThreadStart(targetMethod));
			t.SetApartmentState(ApartmentState.MTA);
			t.Name = "Virtual Server Interop Worker Thread";
			t.IsBackground = true;
			t.Start();
			t.Join();
		}

		internal static ReturnType InvokeOnMtaThread<ReturnType>(NoArgReturningDelegate<ReturnType> targetMethod)
		{
			ThreadStartState<ReturnType> state =
				new ThreadStartState<ReturnType>(
					targetMethod,
					null);

			InvokeOnMtaThread(state);

			if (((IThreadStartState)state).Result is Exception) throw (Exception)((IThreadStartState)state).Result;

			return state.Result;
		}

		internal static ReturnType InvokeOnMtaThread<ReturnType, ArgType>(
				ArgType arg,
				SingleArgReturningDelegate<ReturnType, ArgType> targetMethod)
		{
			ThreadStartState<ReturnType> state =
				new ThreadStartState<ReturnType>(
					targetMethod,
					arg);

			InvokeOnMtaThread(state);

			if (((IThreadStartState)state).Result is Exception)
			{
				throw (Exception)((IThreadStartState)state).Result;
			}
			return state.Result;
		}

		internal static ReturnType InvokeOnMtaThread<ReturnType, ArgOneType, ArgTwoType>(
				ArgOneType arg1,
				ArgTwoType arg2,
				TwoArgReturningDelegate<ReturnType, ArgOneType, ArgTwoType> targetMethod)
		{
			ThreadStartState<ReturnType> state =
				new ThreadStartState<ReturnType>(
					targetMethod,
					new object[] { arg1, arg2 });

			InvokeOnMtaThread(state);

			if (((IThreadStartState)state).Result is Exception) throw (Exception)((IThreadStartState)state).Result;

			return state.Result;
		}

		internal static ReturnType InvokeOnMtaThread<ReturnType, ArgOneType, ArgTwoType, ArgThreeType>(
			ArgOneType arg1,
			ArgTwoType arg2,
			ArgThreeType arg3,
			ThreeArgReturningDelegate<ReturnType, ArgOneType, ArgTwoType, ArgThreeType> targetMethod)
		{
			ThreadStartState<ReturnType> state =
				new ThreadStartState<ReturnType>(
					targetMethod,
					new object[] { arg1, arg2, arg3 });

			InvokeOnMtaThread(state);

			if (((IThreadStartState)state).Result is Exception) throw (Exception)((IThreadStartState)state).Result;

			return state.Result;
		}
		internal static void InvokeOnMtaThread<ArgOneType>(
			ArgOneType a1,
			SingleArgActionDelegate<ArgOneType> targetMethod)
		{
			NoReturnThreadStartState state =
				new NoReturnThreadStartState(
					targetMethod,
					new object[] { a1 });

			InvokeOnMtaThread(state);

			if (((IThreadStartState)state).Result is Exception) throw (Exception)((IThreadStartState)state).Result;

			return; // there is no return value to retrieve
		}

        private static void InvokeOnMtaThread(ThreadStartStateBase state)
        {
            lock (typeof(MtaThreadUtil))
            {
                // TODO: Find a way to reuse the thread

                //Thread t = new Thread(new ParameterizedThreadStart(Invoke));
                //t.SetApartmentState(ApartmentState.MTA);
                //t.Name = "Virtual Server Interop Worker Thread";
                //t.IsBackground = true;
                //t.Start(state);
                //t.Join();

                lock (mtaInvokeRequests)
                {
                    mtaInvokeRequests.Enqueue(state);
                }

                newWorkWaitHandle.Set();

                WaitHandle.WaitAny(new WaitHandle[] { state.MyWaitHandle });
            }
        }
		static MtaThreadUtil()
		{
			Thread t = new Thread(new ThreadStart(FindAndDoWork));
			t.SetApartmentState(ApartmentState.MTA);
			t.Name = "Virtual Server Interop Worker Thread";
			t.IsBackground = true;
			t.Start();
		}
		private static AutoResetEvent newWorkWaitHandle = new AutoResetEvent(false);
		private static Queue<IThreadStartState> mtaInvokeRequests = new Queue<IThreadStartState>();
		private static void FindAndDoWork()
		{
			while (true)
			{
				WaitHandle.WaitAny(new WaitHandle[] { newWorkWaitHandle });

				if (mtaInvokeRequests != null)
				{
					IThreadStartState target = null;
					lock (mtaInvokeRequests)
					{
						if (mtaInvokeRequests.Count > 0)
						{
							target = mtaInvokeRequests.Dequeue();
						}
					}
					if (target != null)
					{
						Invoke(target);
						target.MyWaitHandle.Set();
					}
				}
			}
		}
		private static void Invoke(object state)
		{
			IThreadStartState s = state as IThreadStartState;

			Delegate d = null;
			if (s.TargetObject == null)
			{
				// target method is static
				d = Delegate.CreateDelegate(s.DelegateType, s.TargetMethod);
			}
			else
			{
				// target is an instance method
				d = Delegate.CreateDelegate(s.DelegateType, s.TargetObject, s.TargetMethod.Name, false);
			}

			try
			{
				s.Result = d.Method.Invoke(d.Target, s.Arguments);
			}
			catch (TargetInvocationException ex)
			{
				s.Result = ex.InnerException;
			}
		}
		#endregion
		
		#region ...ThreadStartState...
		private interface IThreadStartState
		{
			Type DelegateType
			{
				get;
				set;
			}
			object TargetObject
			{
				get;
				set;
			}
			MethodInfo TargetMethod
			{
				get;
				set;
			}
			object[] Arguments
			{
				get;
				set;
			}
			object Result
			{
				get;
				set;
			}
			AutoResetEvent MyWaitHandle
			{
				get;
			}
		}
		private abstract class ThreadStartStateBase
			: IThreadStartState
		{
			#region MyWaitHandle
			public AutoResetEvent _MyWaitHandle = new AutoResetEvent(false);
			public AutoResetEvent MyWaitHandle
			{
				get { return this._MyWaitHandle; }
			}
			#endregion
			#region DelegateType
			private Type _DelegateType;
			public Type DelegateType
			{
				get
				{
					return _DelegateType;
				}
				set
				{
					_DelegateType = value;
				}
			}
			#endregion
			#region TargetObject
			private object _TargetObject;
			public object TargetObject
			{
				get
				{
					return _TargetObject;
				}
				set
				{
					_TargetObject = value;
				}
			}
			#endregion
			#region TargetMethod
			private MethodInfo _TargetMethod;
			public MethodInfo TargetMethod
			{
				get
				{
					return _TargetMethod;
				}
				set
				{
					_TargetMethod = value;
				}
			}
			#endregion
			#region Arguments
			private object[] _Arguments;
			public object[] Arguments
			{
				get
				{
					return this._Arguments;
				}
				set
				{
					this._Arguments = value;
				}
			}
			#endregion
			#region Result
			protected object _Result;
			object IThreadStartState.Result
			{
				get
				{
					return this._Result;
				}
				set
				{
					this._Result = value;
				}
			}
			#endregion
			protected ThreadStartStateBase(Delegate d, params object[] args)
			{
				this.DelegateType = d.GetType();
				this.TargetObject = d.Target;
				this.TargetMethod = d.Method;
				this.Arguments = args;
				this._Result = null;
			}
		}
		private class ThreadStartState<ReturnType>
			: ThreadStartStateBase
		{
			#region Result
			public ReturnType Result
			{
				get
				{
					if (_Result is Exception) throw (Exception)_Result;
					return (ReturnType)_Result;
				}
			}
			#endregion
			public ThreadStartState(Delegate d, params object[] args)
				: base(d, args)
			{
			}
		}
		private class NoReturnThreadStartState
			: ThreadStartStateBase
		{
			public NoReturnThreadStartState(Delegate d, params object[] args)
				: base(d, args)
			{
			}
		}
		#endregion

		#region GetMtaProperty
		internal static ReturnType GetMtaProperty<ReturnType>(object target, string propertyName)
		{
			return
				InvokeOnMtaThread<ReturnType, object, string>(
					target,
					propertyName,
					new TwoArgReturningDelegate<ReturnType, object, string>(GetProperty<ReturnType>));
		}

		private static ReturnType GetProperty<ReturnType>(object target, string propertyName)
		{
			PropertyInfo prop = target.GetType().GetProperty(propertyName);
			object val = prop.GetValue(target, null);
			return (ReturnType)val;
		}
		#endregion
		#region SetMtaProperty
		internal static T SetMtaProperty<T>(object target, string propertyName, T newValue)
		{
			// returns T to match pattern of methods

			return
				InvokeOnMtaThread<T, object, string, T>(
					target,
					propertyName,
					newValue,
					new ThreeArgReturningDelegate<T, object, string, T>(SetProperty<T>));
		}
		private static T SetProperty<T>(object target, string propertyName, T newValue)
		{
			PropertyInfo prop = target.GetType().GetProperty(propertyName);
			prop.SetValue(target, newValue, null);
			return newValue;
		}
		#endregion

		#region EnforceMTA
		/// <summary>
		/// Confirms the calling thread has its ApartmentState set to MTA.
		/// </summary>
		internal static void EnforceMTA()
		{
			if (Thread.CurrentThread.GetApartmentState() != ApartmentState.MTA)
			{
				StackFrame sf = new StackFrame(1);
				string methodName = sf.GetMethod().Name;
				if (methodName == ".ctor")
				{
					methodName = "the constructor of " + sf.GetMethod().DeclaringType.Name;
				}
				string msg = "This method (" + methodName + ") requires the calling thread to have a Multi Threaded Apartment (MTA) ApartmentState.";
				
				throw new InvalidOperationException(msg);
			}
		}
		#endregion
		#region EnforceSTA
		/// <summary>
		/// Enforcing STA is to help with testing the API so we don't accidentally miss a call site but
		/// in production no harm would be done if an MTA calls the STA interfaces; it would just "cost"
		/// a little more to call into the API.
		/// </summary>
		// [Conditional("DEBUG")] // this matters now, especially while we're still figuring out
		internal static void EnforceSTA()
		{
			if (Thread.CurrentThread.GetApartmentState() != ApartmentState.STA)
			{
				StackFrame sf = new StackFrame(1);
				string methodName = sf.GetMethod().Name;
				if (methodName == ".ctor")
				{
					methodName = "the constructor of " + sf.GetMethod().DeclaringType.Name;
				}
				string msg = "This method (" + methodName + ") requires the calling thread to have a Single Threaded Apartment (STA) ApartmentState.";

				throw new InvalidOperationException(msg);
			}
		}
		#endregion
	}
}