using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.VirtualServer.Interop;
using System.Threading;
using Bliss.Threading;

namespace Bliss.VirtualServer.MTA
{
	[Author("Anton Delsink", "anton@delsink.com")]
	public sealed class TaskMtaProxy : ITask
	{
		private readonly VMTask task;
		
		#region VM
		private readonly VirtualMachineMtaProxy _VM;
		public IVirtualMachine VM
		{
			get { return _VM; }
		}
		#endregion

		#region VS
		private readonly VirtualServerMtaProxy _VS;
		public IVirtualServer VS
		{
			get { return _VS; }
		}
		#endregion

		#region Constructor
		internal TaskMtaProxy(VirtualMachineMtaProxy vm, VMTask task)
			: this(task)
		{
			this._VM = vm;
			this._VS = null; // to be null means the task is a VM-specific task.
		}
		internal TaskMtaProxy(VirtualServerMtaProxy vs, VMTask task)
			: this(task)
		{
			this._VM = null; // to be null means the task is a Server-specific task.
			this._VS = vs;
		}
		private TaskMtaProxy(VMTask task)
		{
			if (task == null) throw new ArgumentNullException("task");
			this.task = task;
		}
		#endregion

		#region Cancel
		public bool IsCancelable
		{
			get
			{
				MtaThreadUtil.EnforceMTA();

				return this.task.IsCancelable;
			}
		}
		public void Cancel()
		{
			MtaThreadUtil.EnforceMTA();

			if (!this.task.IsCancelable)
			{
				throw new NotSupportedException("This task does not support the Cancel operation.");
			}
			else
			{
				this.task.Cancel();
			}
		}
		#endregion

		#region WaitForCompletion
		public void WaitForCompletion(int timeout)
		{
			this.task.WaitForCompletion(timeout);
		}
		#endregion

		#region Result
		public ITaskResult Result
		{
			get
			{
				return new TaskResultMtaProxy(this.task.Result);
			}
		}
		ITaskResult ITask.Result
		{
			get
			{
				throw new NotImplementedException();
			}
		}
		#endregion

		#region Progress
		private int lastProgress = -1;
		internal bool progressChanged = false;
		public bool ProgressChanged { get { return progressChanged; } }
		public int Progress
		{
			get
			{
				// TODO: Vfy that this will never be negative
				int latestProgress = this.task.PercentCompleted;

				progressChanged = (latestProgress != lastProgress);

				lastProgress = latestProgress;
				return latestProgress;
			}
		}
		#endregion

		#region Description
		public string Description
		{
			get
			{
				return this.task.Description;
			}
		}
		#endregion

		#region IsComplete
		public bool IsComplete
		{
			get
			{
				return this.task.IsComplete;
			}
		}
		#endregion

		public override string ToString()
		{
			MtaThreadUtil.EnforceMTA();

			return this.Description + 
				(this.IsComplete ? " (Completed)" : string.Empty);
		}
	}
}