using System;
using System.Text;
using System.Drawing;
using System.Collections.Generic;
using Microsoft.VirtualServer.Interop;
using System.Threading;
using System.IO;
using Bliss.Threading;

namespace Bliss.VirtualServer.MTA
{
	[Author("Anton Delsink", "anton@delsink.com")]
	public sealed class VirtualMachineMtaProxy : IVirtualMachine
	{
		private readonly IVMVirtualMachine VM;

		#region VS
		private readonly VirtualServerMtaProxy _VS;
		public IVirtualServer VS
		{
			get { return _VS; }
		}
		#endregion

		#region Constructor
		/// <summary>
		/// Represents a configured virtual machine on a specific host (Virtual Server installation).
		/// </summary>
		/// <remarks>
		/// Constructed by a VirtualServerHost instance and thus marked as internal in stead of private.
		/// </remarks>
		/// <param name="vm">COM object representing VM, usually provided by a VirtualMachineHost instance.</param>
		internal VirtualMachineMtaProxy(VirtualServerMtaProxy vs, IVMVirtualMachine vm)
		{
			MtaThreadUtil.EnforceMTA();

			this.VM = vm;
			this._VS = vs;
		}
		#endregion

		#region Properties (with simple return types)
		#region ConfigurationFilePath (.vmc)
		/// <summary>
		/// Full filename of .vmc file (xml configuration file) for the Virtual Machine.
		/// </summary>
		public string ConfigurationFilePath
		{
			get
			{
				MtaThreadUtil.EnforceMTA();

				return this.VM.File;
			}
		}
		#endregion
		#region State
		public VirtualMachineState State
		{
			get
			{
				MtaThreadUtil.EnforceMTA();

				return (VirtualMachineState)(int)this.VM.State;
			}
		}
		#endregion
		#region Name
		public string Name
		{
			get
			{
				MtaThreadUtil.EnforceMTA();

				return this.VM.Name;
			}
		}
		#endregion
		#region Memory
		public int Memory
		{
			get
			{
				MtaThreadUtil.EnforceMTA();

				return this.VM.Memory;
			}
			set
			{
				MtaThreadUtil.EnforceMTA();

				this.VM.Memory = value;
			}
		}
		#endregion
		#region Undoable
		public bool Undoable
		{
			get
			{
				MtaThreadUtil.EnforceMTA();

				return this.VM.Undoable;
			}
			set
			{
				MtaThreadUtil.EnforceMTA();

				this.VM.Undoable = value;
			}
		}
		#endregion
		#region Thumbnail
		/// <summary>
		/// Picture (low res screencapture) of the Virtual Machine's console (screen/monitor) at 64 x 48 pixels.
		/// </summary>
		/// <remarks>
		/// This is really slow. I'm assuming it is the _GenerateThumbnail that is
		/// taking the time and not the array to raster conversion (not my code that)
		/// so the UI will have to deal with refreshing / loading in the background.
		/// </remarks>
		public Bitmap Thumbnail
		{
			get
			{
				MtaThreadUtil.EnforceMTA();

				uint[] thumbnailImage = new uint[64 * 48];
				this.VM.Display._GenerateThumbnail(thumbnailImage);

				// below code from a blog
				Bitmap pic = new Bitmap(64, 48);
				for (int ix = 0; ix < 64 * 48; ix++)
				{
					uint rawColor = thumbnailImage[ix];
					int blue = (int)((rawColor >> 24) % 256);
					int green = (int)((rawColor >> 16) % 256);
					int red = (int)((rawColor >> 8) % 256);

					int row = ix / 64;
					int col = ix % 64;

					pic.SetPixel(col, row, Color.FromArgb(red, green, blue));
				}
				return pic;
			}
		}
		#endregion
		#endregion

		#region HardDiskConnections
		public List<IHardDiskConnection> HardDiskConnections
		{
			get
			{
				MtaThreadUtil.EnforceMTA();

				List<IHardDiskConnection> result = new List<IHardDiskConnection>();
				foreach (VMHardDiskConnection conn in this.VM.HardDiskConnections)
				{
					result.Add(new HardDiskConnectionMtaProxy(conn));
				}
				return result;
			}
		}
		public void AddHardDiskConnection(string fullFileName, DriveBusType bt, int busNumber, int deviceNumber)
		{
			MtaThreadUtil.EnforceMTA();

			this.VM.AddHardDiskConnection(fullFileName, (VMDriveBusType)bt, busNumber, deviceNumber);
		}
		#endregion

		#region DvdDrives
		public List<IDvdDrive> DvdDrives
		{
			get
			{
				MtaThreadUtil.EnforceMTA();

				List<IDvdDrive> result = new List<IDvdDrive>();
				foreach (VMDVDDrive dvd in this.VM.DVDROMDrives)
				{
					result.Add(new DvdDriveMtaProxy(dvd));
				}
				return result;
			}
		}
		#endregion

		#region State Change Methods
		public ITask Start()
		{
			MtaThreadUtil.EnforceMTA();

			return new TaskMtaProxy(this, this.VM.Startup());
		}
		public ITask SaveState()
		{
			MtaThreadUtil.EnforceMTA();

			return new TaskMtaProxy(this, this.VM.Save());
		}
		public ITask TurnOff()
		{
			MtaThreadUtil.EnforceMTA();

			return new TaskMtaProxy(this, this.VM.TurnOff());
		}
		public void DiscardSavedState()
		{
			MtaThreadUtil.EnforceMTA();

			this.VM.DiscardSavedState();
		}
		public void DiscardUndoDisks()
		{
			MtaThreadUtil.EnforceMTA();

			this.VM.DiscardUndoDisks();
		}
		public ITask MergeUndoDisks()
		{
			MtaThreadUtil.EnforceMTA();

			return new TaskMtaProxy(this, this.VM.MergeUndoDisks());
		}
		public void Pause()
		{
			MtaThreadUtil.EnforceMTA();

			this.VM.Pause();
		}
		public ITask Reset()
		{
			MtaThreadUtil.EnforceMTA();

			return new TaskMtaProxy(this, this.VM.Reset());
		}
		public void Resume()
		{
			MtaThreadUtil.EnforceMTA();

			this.VM.Resume();
		}
		#endregion	

        public IVirtualMachine Clone()
        {
            return this.Clone(string.Empty);
        }
        public IVirtualMachine Clone(string diffHardDiskOutputFolder)
        {
            MtaThreadUtil.EnforceMTA();

            // create new VM

            string newVirtualMachinePath = Path.GetDirectoryName(this.VM.File);
            string newVirtualMachineName = this.Name + " Clone";
            
            // is the new VM name unique?
            foreach (IVirtualMachine vm in this.VS.VirtualMachines)
            {
                if (vm.Name.ToUpper() == newVirtualMachineName.ToUpper())
                {
                    newVirtualMachineName += " " + Guid.NewGuid().ToString();
                    break;
                }
            }

            IVirtualMachine newVM = this.VS.CreateVirtualMachine(newVirtualMachineName, newVirtualMachinePath);
            newVM.Memory = this.Memory;

            // create diff disks
            foreach (HardDiskConnectionMtaProxy conn in this.HardDiskConnections)
            {
                string baseDiskFileName = conn.VirtualHardDisk.File;
                string newDiskFileName = 
                    (string.IsNullOrEmpty(diffHardDiskOutputFolder) ? Path.GetDirectoryName(baseDiskFileName) : diffHardDiskOutputFolder) +
                    "\\" +
                    Path.GetFileName(baseDiskFileName)
                        .Replace(".vhd", ".diff.vhd");

                ITask t = this.VS.CreateDifferencingVirtualHardDisk(newDiskFileName, baseDiskFileName);
                t.WaitForCompletion(-1);

                //FileAttributes faBefore = File.GetAttributes(newDiskFileName);
                //FileAttributes faAfter = faBefore & ((FileAttributes)int.MaxValue ^ FileAttributes.Compressed);
                //File.SetAttributes(newDiskFileName, faAfter);

                newVM.AddHardDiskConnection(newDiskFileName, conn.BusType, conn.BusNumber, conn.DeviceNumber);
            }

            // TODO: Configure networks in the same way

            return newVM;
        }

        public override string ToString()
        {
            return this.Name;
        }
	}
}
