using System;
using System.Text;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using Microsoft.VirtualServer.Interop;
using System.Collections.ObjectModel;
using System.Threading;
using Bliss.Threading;

namespace Bliss.VirtualServer.MTA
{
	[Author("Anton Delsink", "anton@delsink.com")]
	public sealed class VirtualServerMtaProxy : IVirtualServer
	{
		private readonly IVMVirtualServer VS; // target VS installation; the host being managed; COM interop here!

		#region Constructor
		internal VirtualServerMtaProxy(VirtualServerManager vsm, string virtualServerHostName)
		{
			MtaThreadUtil.EnforceMTA();

			this._VirtualServerManager = vsm;
			this.VS = InitVS.GetVMVirtualServerClass(virtualServerHostName);
		}
		internal static VirtualServerMtaProxy Create(VirtualServerManager vsm, string virtualServerHostName)
		{
			MtaThreadUtil.EnforceMTA();

			return new VirtualServerMtaProxy(vsm, virtualServerHostName);
		}
		#endregion

		#region Properties
		#region VirtualMachines
		public ReadOnlyCollection<IVirtualMachine> VirtualMachines
		{
			get
			{
				MtaThreadUtil.EnforceMTA();

				return new ReadOnlyCollection<IVirtualMachine>(
					GetVirtualMachines(this));
			}
		}
		#region GetVirtualMachines
		private static List<IVirtualMachine> GetVirtualMachines(VirtualServerMtaProxy targetVSH)
		{
			MtaThreadUtil.EnforceMTA();

			List<IVirtualMachine> vmList = new List<IVirtualMachine>();
			foreach (IVMVirtualMachine vm in targetVSH.VS.VirtualMachines)
			{
				vmList.Add(new VirtualMachineMtaProxy(targetVSH, vm));
			}
			return vmList;
		}
		#endregion
		#endregion
		#region VirtualServerManager
		private VirtualServerManager _VirtualServerManager;
		public VirtualServerManager VirtualServerManager
		{
			get
			{
				return this._VirtualServerManager;
			}
			set
			{
				this._VirtualServerManager = value;
			}
		}
		#endregion
        #region HostName
        public string HostName
        {
            get
            {
                MtaThreadUtil.EnforceMTA();
                return this.VirtualServerManager.VirtualServerHostName;
            }
        }
        #endregion
        #region SearchFolders
        public string[] SearchFolders
        {
            get
            {
                MtaThreadUtil.EnforceMTA();

                List<string> folderList = new List<string>();
                foreach (object folder in (object[])this.VS.SearchPaths)
                {
                    folderList.Add(folder.ToString());
                }
                return folderList.ToArray();
            }
        }
        #endregion
        #endregion

        public void AddSearchFolder(string targetPath)
        {
            MtaThreadUtil.EnforceMTA();

            object[] currentSearchFolders = (object[])this.VS.SearchPaths;
            string[] newSearchFolders = new string[currentSearchFolders.Length + 1];
            Array.Copy(currentSearchFolders, newSearchFolders, currentSearchFolders.Length);
            newSearchFolders[newSearchFolders.Length - 1] = targetPath;

            this.VS.SearchPaths = newSearchFolders;
        }
        public IVirtualMachine AddExistingVMC(string fullFileName)
		{
			MtaThreadUtil.EnforceMTA();

			return new VirtualMachineMtaProxy(this, this.VS.RegisterVirtualMachine(null, fullFileName));
		}
		public IVirtualMachine CreateVirtualMachine(string vmName, string vmcFolder)
		{
			MtaThreadUtil.EnforceMTA();

			return new VirtualMachineMtaProxy(this, this.VS.CreateVirtualMachine(vmName, vmcFolder));
		}
		public ITask CreateDifferencingVirtualHardDisk(string fullFileNameForNewVHD, string fullFileNameForBaseVHD)
		{
			MtaThreadUtil.EnforceMTA();

			TaskMtaProxy t = new TaskMtaProxy(this,
				this.VS.CreateDifferencingVirtualHardDisk(
					fullFileNameForNewVHD,
					fullFileNameForBaseVHD));

			return t;
		}
        public string GetBaseDiskFileName(string targetVhdFileName)
        {
            MtaThreadUtil.EnforceMTA();

            VMHardDisk vhd = this.VS.GetHardDisk(targetVhdFileName);
            return vhd.Parent == null ? string.Empty : vhd.Parent.File;
        }
	}
}