using System;
using System.Text;
using System.Drawing;
using System.Collections.Generic;
using Microsoft.VirtualServer.Interop;

namespace DTL.VirtualServer
{
	public sealed class VirtualMachine
	{
		private readonly IVMVirtualMachine VM;
		#region VS
		private readonly VirtualServerHost _VS;
		public VirtualServerHost VS
		{
			get { return _VS; }
		}
		#endregion

		#region Constructor
		/// <summary>
		/// Represents a configured virtual machine on a specific host (Virtual Server installation).
		/// </summary>
		/// <remarks>
		/// Constructed by a VirtualServerHost instance and thus marked as internal in stead of private.
		/// </remarks>
		/// <param name="vm">COM object representing VM, usually provided by a VirtualMachineHost instance.</param>
		internal VirtualMachine(VirtualServerHost vs, IVMVirtualMachine vm)
		{
			this.VM = vm;
			this._VS = vs;
		}
		#endregion

		#region HardDiskConnections
		public List<HardDiskConnection> HardDiskConnections
		{
			get
			{
				List<HardDiskConnection> result = new List<HardDiskConnection>();
				foreach (VMHardDiskConnection conn in this.VM.HardDiskConnections)
				{
					result.Add(new HardDiskConnection(conn));
				}
				return result;
			}
		}
		public void AddHardDiskConnection(string fullFileName, DriveBusType bt, int busNumber, int deviceNumber)
		{
			this.VM.AddHardDiskConnection(fullFileName, (VMDriveBusType)bt, busNumber, deviceNumber);
		}
		#endregion

		#region Properties
		#region ConfigurationFilePath (.vmc)
		/// <summary>
		/// Full filename of .vmc file (xml configuration file) for the Virtual Machine.
		/// </summary>
		public string ConfigurationFilePath
		{
			get
			{
				return this.VM.File;
			}
		}
		#endregion
		#region State
		public VirtualMachineState State
		{
			get
			{
				return (VirtualMachineState)(int)this.VM.State;
			}
		}
		#endregion
		#region Name
		public string Name
		{
			get
			{
				return this.VM.Name;
			}
		}
		#endregion
		#region Memory
		public int Memory
		{
			get
			{
				return this.VM.Memory;
			}
			set
			{
				this.VM.Memory = value;
			}
		}
		#endregion
		#region Undoable
		public bool Undoable
		{
			get
			{
				return this.VM.Undoable;
			}
			set
			{
				this.VM.Undoable = value;
			}
		}
		#endregion
		#region Thumbnail
		/// <summary>
		/// Picture (low res screencapture) of the Virtual Machine's console (screen/monitor) at 64 x 48 pixels.
		/// </summary>
		/// <remarks>
		/// This is really slow. I'm assuming it is the _GenerateThumbnail that is
		/// taking the time and not the array to raster conversion (not my code that)
		/// so the UI will have to deal with refreshing / loading in the background.
		/// </remarks>
		public Bitmap Thumbnail
		{
			get
			{
				uint[] thumbnailImage = new uint[64 * 48];
				this.VM.Display._GenerateThumbnail(thumbnailImage);

				// below code from a blog
				Bitmap pic = new Bitmap(64, 48);
				for (int ix = 0; ix < 64 * 48; ix++)
				{
					uint rawColor = thumbnailImage[ix];
					int blue = (int)((rawColor >> 24) % 256);
					int green = (int)((rawColor >> 16) % 256);
					int red = (int)((rawColor >> 8) % 256);

					int row = ix / 64;
					int col = ix % 64;

					pic.SetPixel(col, row, Color.FromArgb(red, green, blue));
				}
				return pic;
			}
		}
		#endregion
		#endregion

		#region State Change Methods
		public Task Start()
		{
			return new Task(this, this.VM.Startup());
		}
		public Task SaveState()
		{
			return new Task(this, this.VM.Save());
		}
		public Task TurnOff()
		{
			return new Task(this, this.VM.TurnOff());
		}
		public void DiscardSavedState()
		{
			this.VM.DiscardSavedState();
		}
		public void DiscardUndoDisks()
		{
			this.VM.DiscardUndoDisks();
		}
		public Task MergeUndoDisks()
		{
			return new Task(this, this.VM.MergeUndoDisks());
		}
		public void Pause()
		{
			this.VM.Pause();
		}
		public Task Reset()
		{
			return new Task(this, this.VM.Reset());
		}
		public void Resume()
		{
			this.VM.Resume();
		}
		#endregion
	}
}
