using System;
using System.Text;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using Microsoft.VirtualServer.Interop;
using System.Collections.ObjectModel;
using System.Threading;

namespace DTL.VirtualServer
{
	public sealed partial class VirtualServerHost : IDisposable
	{
		private readonly IVMVirtualServer VS; // target VS installation; the host being managed.

		private System.Threading.Timer statusMonitor; // initialized in Constructor
		
		#region Constructors
		/// <summary>
		/// Connects to the local installation of Virtual Server R2 by connecting to the network name "localhost".
		/// </summary>
		public VirtualServerHost()
			: this("localhost")
		{
		}
		/// <summary>
		/// Connects to the specified Virtual Server R2 installation.
		/// </summary>
		/// <param name="targetServerName">The network name of the host running Virtual Server R2 (DNS or NetBIOS name).</param>
		public VirtualServerHost(string targetServerName)
		{
			try
			{
				InitVS vsInitHelper = new InitVS();

				//System.Threading.Thread.CurrentThread.SetApartmentState(System.Threading.ApartmentState.STA);

				this.VS = vsInitHelper.GetVMVirtualServerClass(targetServerName);

				this.statusMonitor = new System.Threading.Timer(
					new System.Threading.TimerCallback(StatusCheck_Tick),
					null,
					1000,
					250); // TODO: Make configurable
			}
			catch (COMException ex)
			{
				throw new NotSupportedException("To use the DTL.VirtualServer library in Visual Studio 2005's debug mode, please disable the Visual Studio Hosting Process. Alternatively, you may have the [STAThread] attribute in place; it is not supported when using this library.", ex);
			}
			catch (UnauthorizedAccessException ex)
			{
				throw new ApplicationException("Check that the Virtual Server is started and that the account used to access it has the appropriate permissions.", ex);
			}
			catch (Exception)
			{
				throw; // we don't get a good reason (for like when the server in unavailable) so we pass the dumb msg up the stack
			}
		}
		#endregion

		#region Tasks

		//public delegate void TaskCompletedDelegate(object sender, TaskCompletedEventArgs e);
		public event EventHandler<TaskCompletedEventArgs> TaskCompleted;

		//public delegate void TaskProgressChangedDelegate(object sender, TaskProgressChangedEventArgs e);
		public event EventHandler<TaskProgressChangedEventArgs> TaskProgressChanged;

		/// <summary>
		/// Complete list of all known pending tasks.
		/// </summary>
		/// <remarks>
		/// The VirtualServerHost instance is responsible for maintaining this list. The
		/// Task objects are not aware of this list and do not know their parent host.
		/// </remarks>
		private List<Task> tasks = new List<Task>();

		/// <summary>
		/// Status indicators and descriptions for requested actions currently still being performed by the Virtual Server host.
		/// </summary>
		public ReadOnlyCollection<Task> ActiveTasks
		{
			get
			{
				List<Task> activeTasks = new List<Task>();
				lock (tasks)
				{
					foreach (Task t in tasks)
					{
						if (!t.IsComplete)
						{
							activeTasks.Add(t);
						}
					}
				}
				return new ReadOnlyCollection<Task>(activeTasks);
			}
		}

		internal void RememberTask(Task task)
		{
			lock (tasks)
			{
				tasks.Add(task);
			}
		}

		private void StatusCheck_Tick(object state)
		{
			List<Task> completedTasks = new List<Task>();
			List<Task> progressChangedTasks = new List<Task>();

			lock (tasks)
			{
				foreach (Task t in tasks)
				{
					if (t.IsComplete) completedTasks.Add(t);
					if (t.progressChanged) progressChangedTasks.Add(t);
				}
			}
			
			// TODO: TryCatch so that evil event handlers don't DoS the removal of completed tasks
			
			foreach (Task t in completedTasks)
			{
				OnTaskCompleted(t);
				lock (this.tasks)
				{
					this.tasks.Remove(t);
				}
			}

			foreach (Task t in progressChangedTasks)
			{
				OnTaskProgressChanged(t);
			}
		}
		private void OnTaskCompleted(Task t)
		{
			// TODO: Make into a Component so that we can know the Form and thus the main UI thread
			// INF: Raising even on backgroud thread!
			TaskCompleted(this, new TaskCompletedEventArgs(t));
		}
		private void OnTaskProgressChanged(Task t)
		{
			// TODO: Make into a Component so that we can know the Form and thus the main UI thread
			// INF: Raising even on backgroud thread!
			TaskProgressChanged(this, new TaskProgressChangedEventArgs(t));
		}
		#endregion

		#region VirtualMachines

		public List<VirtualMachine> VirtualMachines
		{
			get
			{
				List<VirtualMachine> vmList = new List<VirtualMachine>();
				foreach (IVMVirtualMachine vm in this.VS.VirtualMachines)
				{
					vmList.Add(new VirtualMachine(this, vm));
				}
				return vmList;
			}
		}

		public VirtualMachine AddExistingVMC(string fullFileName)
		{
			return new VirtualMachine(this, this.VS.RegisterVirtualMachine(null, fullFileName));
		}

		public VirtualMachine CreateVirtualMachine(string vmName, string vmcFolder)
		{
			return new VirtualMachine(this, this.VS.CreateVirtualMachine(vmName, vmcFolder));
		}
		#endregion

		#region VHDs
		public Task CreateDifferencingVirtualHardDisk(string fullFileNameForNewVHD, string fullFileNameForBaseVHD)
		{
			Task t = new Task(this,
				this.VS.CreateDifferencingVirtualHardDisk(
					fullFileNameForNewVHD, fullFileNameForBaseVHD));

			this.RememberTask(t);

			return t;
		}
		#endregion

		#region IDisposable Members

		public void Dispose()
		{
			this.statusMonitor.Dispose();
		}

		#endregion

		public static void InvokeOnMtaThread(ThreadStart ts)
		{

		}
	}
}