using System;
using System.Text;
using System.Drawing;
using System.Collections.Generic;
using Microsoft.VirtualServer.Interop;
using System.Threading;
using Delsink.VirtualServer.MTA;
using Delsink.Threading;

namespace Delsink.VirtualServer.STA
{
	[Author("Anton Delsink", "anton@delsink.com")]
	public class VirtualMachineStaProxy : IVirtualMachine
	{
		private VirtualMachineMtaProxy proxy;

		#region VS
		private VirtualServerStaProxy _VS;
		public IVirtualServer VS
		{
			get
		    {
				return this._VS;
		    }
		}
		#endregion

		#region Constructor
		internal VirtualMachineStaProxy(VirtualServerStaProxy vs, IVirtualMachine proxy)
		{
			if (!(proxy is VirtualMachineMtaProxy)) throw new ArgumentException();

			this._VS = vs;
			this.proxy = (VirtualMachineMtaProxy)proxy;
		}
		#endregion

		#region Properties (with simple return types)
		public string ConfigurationFilePath
		{
			get
			{
				MtaThreadUtil.EnforceSTA();

				return MtaThreadUtil.GetMtaProperty<string>(this.proxy, "ConfigurationFilePath");
			}
		}
		public VirtualMachineState State
		{
			get
			{
				MtaThreadUtil.EnforceSTA();

				return MtaThreadUtil.GetMtaProperty<VirtualMachineState>(this.proxy, "State");
			}
		}
		public string Name
		{
			get
			{
				MtaThreadUtil.EnforceSTA();

				return MtaThreadUtil.GetMtaProperty<string>(this.proxy, "Name");
			}
		}
		public int Memory
		{
			get
			{
				MtaThreadUtil.EnforceSTA();

				return MtaThreadUtil.GetMtaProperty<int>(this.proxy, "Memory");
			}
			set
			{
				MtaThreadUtil.EnforceSTA();

				MtaThreadUtil.SetMtaProperty<int>(this.proxy, "Memory", value);
			}
		}
		public Bitmap Thumbnail
		{
			get
			{
				//MtaThreadUtil.EnforceSTA(); // This is not checked here because the Background Worker is used to get a thumbnail and it doesn't match the STA state of the UI but rather the MTA state of the Main method's thread.

				return MtaThreadUtil.GetMtaProperty<Bitmap>(this.proxy, "Thumbnail");
			}
		}
		public bool Undoable
		{
			get
			{
				MtaThreadUtil.EnforceSTA();

				return MtaThreadUtil.GetMtaProperty<bool>(this.proxy, "Undoable");
			}
			set
			{
				MtaThreadUtil.EnforceSTA();

				MtaThreadUtil.SetMtaProperty<bool>(this.proxy, "Undoable", value);
			}
		}
		#endregion

		#region State Change Methods
		public ITask Start()
		{
			MtaThreadUtil.EnforceSTA();

			MtaThreadUtil.NoArgReturningDelegate<ITask> startup =
				new MtaThreadUtil.NoArgReturningDelegate<ITask>(this.proxy.Start);

			return 
				new TaskStaProxy(
					this,
					MtaThreadUtil.InvokeOnMtaThread<ITask>(
						startup));
		}

		public ITask SaveState()
		{
			MtaThreadUtil.EnforceSTA();

			MtaThreadUtil.NoArgReturningDelegate<ITask> saveState =
				new MtaThreadUtil.NoArgReturningDelegate<ITask>(this.proxy.SaveState);

			return 
				new TaskStaProxy(
					this,
					MtaThreadUtil.InvokeOnMtaThread<ITask>(
						saveState));
		}

		public ITask TurnOff()
		{
			MtaThreadUtil.EnforceSTA();

			MtaThreadUtil.NoArgReturningDelegate<ITask> turnOff =
				new MtaThreadUtil.NoArgReturningDelegate<ITask>(this.proxy.TurnOff);

			return 
				new TaskStaProxy(
					this,
					MtaThreadUtil.InvokeOnMtaThread<ITask>(
						turnOff));
		}
		public void DiscardSavedState()
		{
			MtaThreadUtil.EnforceSTA();

			MtaThreadUtil.InvokeOnMtaThread(new ThreadStart(this.proxy.DiscardSavedState));
		}
		public void DiscardUndoDisks()
		{
			MtaThreadUtil.EnforceSTA();

			MtaThreadUtil.InvokeOnMtaThread(new ThreadStart(this.proxy.DiscardUndoDisks));
		}
		public ITask MergeUndoDisks()
		{
			MtaThreadUtil.EnforceSTA();

			MtaThreadUtil.NoArgReturningDelegate<ITask> merge =
				new MtaThreadUtil.NoArgReturningDelegate<ITask>(this.proxy.MergeUndoDisks);

			return 
				new TaskStaProxy(
					this,
					MtaThreadUtil.InvokeOnMtaThread<ITask>(
						merge));
		}
		public void Pause()
		{
			MtaThreadUtil.EnforceSTA();

			MtaThreadUtil.InvokeOnMtaThread(new ThreadStart(this.proxy.Pause));
		}
		public ITask Reset()
		{
			MtaThreadUtil.EnforceSTA();

			MtaThreadUtil.NoArgReturningDelegate<ITask> reset =
				new MtaThreadUtil.NoArgReturningDelegate<ITask>(this.proxy.Reset);

			return
				new TaskStaProxy(
					this,
					MtaThreadUtil.InvokeOnMtaThread<ITask>(
						reset));
		}
		public void Resume()
		{
			MtaThreadUtil.EnforceSTA();

			MtaThreadUtil.InvokeOnMtaThread(new ThreadStart(this.proxy.Resume));
		}
		#endregion

		#region DvdDrives
		public List<IDvdDrive> DvdDrives
		{
			get
			{
				MtaThreadUtil.EnforceSTA();

				List<IDvdDrive> result = new List<IDvdDrive>();
				foreach (DvdDriveMtaProxy dvd in MtaThreadUtil.GetMtaProperty<List<IDvdDrive>>(this.proxy, "DvdDrives"))
				{
					result.Add(new DvdDriveStaProxy(dvd));
				}
				return result;
			}
		}
		#endregion

		#region HardDiskConnections
		public List<IHardDiskConnection> HardDiskConnections
		{
			get
			{
				MtaThreadUtil.EnforceSTA();

				List<IHardDiskConnection> result = new List<IHardDiskConnection>();
				foreach (HardDiskConnectionMtaProxy conn in MtaThreadUtil.GetMtaProperty<List<IHardDiskConnection>>(this.proxy, "HardDiskConnections"))
				{
					result.Add(new HardDiskConnectionStaProxy(conn));
				}
				return result;
			}
		}
		public void AddHardDiskConnection(string fullFileName, DriveBusType bt, int busNumber, int deviceNumber)
		{
			MtaThreadUtil.EnforceSTA();

			// TODO: Test passing of args
			Thread t = new Thread(delegate()
			{
				this.proxy.AddHardDiskConnection(
					fullFileName,
					bt,
					busNumber,
					deviceNumber);
			});
			t.SetApartmentState(ApartmentState.MTA);
			t.Start();
			t.Join();
		}
		#endregion

        public IVirtualMachine Clone(string diffHardDiskOutputFolder)
        {
            MtaThreadUtil.EnforceSTA();

            return
                new VirtualMachineStaProxy(
                    this.VS as VirtualServerStaProxy,
                    MtaThreadUtil.InvokeOnMtaThread<IVirtualMachine, string>(
                        diffHardDiskOutputFolder,
                        new MtaThreadUtil.SingleArgReturningDelegate<IVirtualMachine, string>(
                            this.proxy.Clone)));
        }
        public IVirtualMachine Clone()
        {
            MtaThreadUtil.EnforceSTA();

            return
                new VirtualMachineStaProxy(
                    this.VS as VirtualServerStaProxy,
                    MtaThreadUtil.InvokeOnMtaThread<IVirtualMachine>(
                        new MtaThreadUtil.NoArgReturningDelegate<IVirtualMachine>(
                            this.proxy.Clone)));
        }

        public override string ToString()
        {
            return this.Name;
        }
	}
}