using System;
using System.ComponentModel;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;

using System.Collections.ObjectModel;
using System.Threading;
using System.Runtime.InteropServices;
using Microsoft.VirtualServer.Interop;

using Delsink.VirtualServer;
using System.Configuration;
using Delsink.VirtualServer.STA;

using Form = System.Windows.Forms.Form;
using Application = System.Windows.Forms.Application;
using Delsink.Threading;

namespace Delsink
{
	/// <summary>
	/// An alternative to the COM API for configuring Microsoft Virtual Server and
	/// its guest Virtual Machines. This Component is for use on .NET Windows Forms
	/// (specifically System.Windows.Forms.Form) as an abstraction, providing
	/// further interfaces from the DTL.VirtualServer namespace for access to specific
	/// configuration options.
	/// </summary>
	/// <remarks>
	/// All code in this class must be willing and able to run on the UI thread which
	/// is extremely likely to be STA (Single Threaded Apartment) for the sake of
	/// COM Interop (for common dialogs, etc.). Consider this component to be a User 
	/// Interface Control like a TextBox: external users call it from an STA Thread,
	/// internally it will make use of the MtaThreadUtil to marshal calls to an 
	/// MTA Thread as and when necessary. When you see an interface (IVirtualServer,
	/// IVirtualMachine, ITask, etc.) then you can access members from an STA thread.
	/// When you're working with DTL.VirtualMachine.DTL classes (VirtualServer,
	/// VirtualMachine, Task, etc.), then the thread has to be MTA.
	/// </remarks>
	[Author("Anton Delsink", "anton@delsink.com")]
	public sealed partial class VirtualServerManager : Component
	{
		#region Initialize
		/// <summary>
		/// Configures the process for COM Interop as required by the
		/// underlying Virtual Server COM API and checks for MTA apartment 
		/// state. The application entry point should call this method as
		/// the very first statement and use an MTAThread to do so.
		/// </summary>
		public static void Initialize()
		{
			MtaThreadUtil.EnforceMTA();

			try
			{
				Delsink.VirtualServer.MTA.InitVS.CoInitializeSecurity(IntPtr.Zero,
					-1,
					IntPtr.Zero,
					IntPtr.Zero,
					(uint)Delsink.VirtualServer.MTA.RpcAuthnLevel.PktPrivacy,
					(uint)Delsink.VirtualServer.MTA.RpcImpLevel.Impersonate, // non-default and required by VS
					IntPtr.Zero,
					(uint)Delsink.VirtualServer.MTA.EoAuthnCap.DynamicCloaking,
					IntPtr.Zero);
			}
			catch (COMException ex)
			{
				throw new InvalidOperationException("CoInitializeSecurity failed. Make sure VirtualServerManager.Initialize is the very first method called in the application and that the thread is MTA.", ex);
			}
		}
		#endregion

		#region Constructors
		/// <summary>
		/// Only exists for Designer support. See overloaded constructor with
		/// argument IContainer.
		/// </summary>
		public VirtualServerManager()
			: this(null)
		{
		}
		/// <summary>
		/// With the VSM placed on a Windows Form (its container), the component
		/// can use the container (a Control) to .Invoke methods on the UI's
		/// thread (important for events).
		/// </summary>
		/// <param name="container">Usually a Form but can be a Control.</param>
		public VirtualServerManager(IContainer container)
		{
			if (container != null) container.Add(this);

			InitializeComponent();
		}
		#endregion

		#region VirtualServer

		private string virtualServerHostName = string.Empty;

		/// <summary>
		/// Each VirtualServerManager refers to and manages exactly one Virtual Server.
		/// </summary>
		private IVirtualServer _VirtualServer = null;

		/// <summary>
		/// The choice of Virtual Server to be managed by this VirtualServerManager is 
		/// effectively Read Only and is not expected to change for the lifetime 
		/// of the this object.
		/// </summary>
		public IVirtualServer VirtualServer
		{
			get
			{
				if (this._VirtualServer == null)
				{
					// TODO: test remote connections, take on security issues in workgroups
					string targetServerName = ConfigurationManager.AppSettings["VirtualServerHostName"];

					// if not configured, default to "localhost" (no app.config for DLLs)
					targetServerName = string.IsNullOrEmpty(targetServerName) ? "localhost" : targetServerName;

					this._VirtualServer = new VirtualServerStaProxy(this, targetServerName);
					this.virtualServerHostName = targetServerName;
				}

				return this._VirtualServer;
			}
		}
		#endregion
		
		#region Tasks

		public event EventHandler<TaskCompletedEventArgs> TaskCompleted;
		public event EventHandler<TaskProgressChangedEventArgs> TaskProgressChanged;
		public event EventHandler<TaskAddedEventArgs> TaskAdded;

		/// <summary>
		/// Complete list of all known pending tasks as issued through this library.
		/// </summary>
		/// <remarks>
		/// The IVirtualServer instance is responsible for maintaining this list. The
		/// Task objects are not aware of this list and do not know their parent IVirtualMachine,
		/// IVirtualMachine or VirtualServerManager.
		/// </remarks>
		private List<ITask> tasks = new List<ITask>();

		/// <summary>
		/// Status indicators and descriptions for requested actions currently still being performed by the Virtual Server host.
		/// </summary>
		/// <remarks>
		/// This list is a selection from the internal tasks collection.
		/// </remarks>
		public ReadOnlyCollection<ITask> ActiveTasks
		{
			get
			{
				List<ITask> activeTasks = new List<ITask>();
				lock (this.tasks)
				{
					foreach (ITask t in this.tasks)
					{
						if (!t.IsComplete)
						{
							activeTasks.Add(t);
						}
					}
				}
				return new ReadOnlyCollection<ITask>(activeTasks);
			}
		}

		/// <summary>
		/// For adding a task to the list of known tasks.
		/// </summary>
		/// <remarks>
		/// This should be done when the command is invoked regardless of
		/// whether the task completed before leaving the command method as
		/// the list will be polled for Completions.
		/// </remarks>
		/// <param name="task">Description of pending / in progress action requested of Virtual Server.</param>
		internal void RememberTask(ITask task)
		{
			if (task == null) return;
			if (!(task is TaskStaProxy)) throw new InvalidOperationException("Only STA Proxies should be used with the VirtualServerManager.");

			lock (this.tasks)
			{
				tasks.Add(task);
			}

			OnTaskAdded(task);

			timerTaskRefresh.Enabled = true;
		}

		/// <summary>
		/// Periodically invoked by the Timer component (taskStatusCheckTimer)
		/// on a background thread. It iterates through the list of tasks
		/// and raises an event as and when progress has changed or a task has
		/// completed.
		/// </summary>
		private void timerTaskRefresh_Tick(object sender, EventArgs e)
		{
			List<ITask> completedTasks = new List<ITask>();
			List<ITask> progressChangedTasks = new List<ITask>();

			lock (tasks)
			{
				foreach (ITask t in tasks)
				{
					if (t.IsComplete) completedTasks.Add(t);
					if (t.ProgressChanged) progressChangedTasks.Add(t);
				}
			}

			// TODO: TryCatch so that evil event handlers don't DoS the removal of completed tasks

			foreach (ITask t in completedTasks)
			{
				OnTaskCompleted(t);
				lock (this.tasks)
				{
					this.tasks.Remove(t);
				}
			}

			foreach (ITask t in progressChangedTasks)
			{
				OnTaskProgressChanged(t);
			}

			lock (this.tasks)
			{
				if (tasks.Count == 0)
				{
					timerTaskRefresh.Enabled = false;
				}
			}
		}
		private void OnTaskAdded(ITask t)
		{
			if (TaskAdded != null)
			{
				RaiseEventOnContainerThread<TaskAddedEventArgs>(
					new EventHandler<TaskAddedEventArgs>(TaskAdded),
					new TaskAddedEventArgs(t));
			}
		}
		private void OnTaskCompleted(ITask t)
		{
			if (TaskCompleted != null) // at least one subscriber to the event
			{
				RaiseEventOnContainerThread<TaskCompletedEventArgs>(
					new EventHandler<TaskCompletedEventArgs>(TaskCompleted),
					new TaskCompletedEventArgs(t));
			}
		}
		private void OnTaskProgressChanged(ITask t)
		{
			if (TaskProgressChanged != null) // at least one subscriber to the event
			{
				RaiseEventOnContainerThread<TaskProgressChangedEventArgs>(
					new EventHandler<TaskProgressChangedEventArgs>(TaskProgressChanged),
					new TaskProgressChangedEventArgs(t));
			}
		}
		private void RaiseEventOnContainerThread<T>(EventHandler<T> meth, T t)
			where T : EventArgs
		{
			// TODO: There has got to be a better way to deal with this!
			if (Application.OpenForms[0] != null)
			{
				Application.OpenForms[0].Invoke(
					meth,
					new object[] { this, t });
			}
		}
		#endregion

		#region VmrcDefaultPath
		private static string defaultVmrcDefaultPath = @"C:\program files\Microsoft Virtual Server\VMRC Client\vmrc.exe";
		private static string _VmrcDefaultPath;
		public static string VmrcDefaultPath
		{
			get
			{
				if (_VmrcDefaultPath == null)
				{
					_VmrcDefaultPath = ConfigurationManager.AppSettings["VmrcDefaultPath"];

					if (string.IsNullOrEmpty(_VmrcDefaultPath))
					{
						_VmrcDefaultPath = defaultVmrcDefaultPath;
					}
				}
				return _VmrcDefaultPath;
			}
			set
			{
				_VmrcDefaultPath = value;
			}
		}
		#endregion
	
		public static void StartVmrc(string virtualMachineName)
		{
			ProcessStartInfo targetEXE = new ProcessStartInfo();
			targetEXE.FileName = VmrcDefaultPath;
			targetEXE.Arguments = virtualMachineName;

			Process proc = Process.Start(targetEXE);
		}
	}
}