using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

using Delsink.VirtualServer.MTA;
using System.IO;
using Delsink.VirtualServer;

namespace Bliss.VPC
{
	/// <summary>
	/// Look-alike Virtual PC Console form.
	/// </summary>
	public partial class VirtualMachineConsoleForm : Form
	{
		/// <summary>
		/// The flow layout panel contains a list of VMUCs, of which one
		/// may be selected at a time. This field refers to the selected
		/// VMUC and thus selected VM.
		/// </summary>
		private VirtualMachineUserControl SelectedVirtualMachineUserControl = null;

		#region Constructor
		public VirtualMachineConsoleForm()
		{
			InitializeComponent();
		}
		#endregion
		
		private void VirtualMachineConsoleForm_Load(object sender, EventArgs e)
		{
			InitVirtualMachineGroupsTree();
			HideVirtualMachineGroupsTree(); // for visibility, but also removed from .Controls

			try
			{
				foreach (IVirtualMachine vm in this.vsm.VirtualServer.VirtualMachines)
				{
					AddVirtualMachineToFlowLayoutPanel(vm);
					AddVirtualMachineToGroupsTree(vm);
				}
			}
			catch (Exception)
			{
				MessageBox.Show("Failed to contact the local Virtual Server for a list of Virtual Machines. Please check that the service is started and that you have the appropriate permissions.");
				this.Visible = false;
				this.Close();
			}
		}

		void ucVM_Selected(object sender, EventArgs e)
		{
			if (SelectedVirtualMachineUserControl != null)
			{
				// causes display properties to reflect controls' selected state
				SelectedVirtualMachineUserControl.IsSelected = false;
			}

			SelectedVirtualMachineUserControl = sender as VirtualMachineUserControl;

			SyncVirtualMachineSelection();
		}

		#region Flow Layout Panel Helpers
		
		private void AddVirtualMachineToFlowLayoutPanel(IVirtualMachine vm)
		{
			VirtualMachineUserControl ucVM = new VirtualMachineUserControl();
			ucVM.VM = vm;
			
			// for sync with tree:
			ucVM.Selected += new VirtualMachineUserControl.VirtualMachineUserControlSelectedDelegate(ucVM_Selected);
			
			flpVMs.Controls.Add(ucVM);
		}


		private void SyncVirtualMachineSelection()
		{
			VirtualMachineTreeNode tnVM = tvGroups.SelectedNode as VirtualMachineTreeNode;
			if (tnVM != null &&
				tnVM.VM == SelectedVirtualMachineUserControl.VM)
			{
				// already in sync
				return;
			}

			foreach (TreeNode topLevelNode in tvGroups.Nodes)
			{
				if (topLevelNode is VirtualMachineTreeNode)
				{
					if (topLevelNode.Text == SelectedVirtualMachineUserControl.VM.Name)
					{
						tvGroups.SelectedNode = topLevelNode;
						topLevelNode.EnsureVisible();
						topLevelNode.TreeView.Select();
						return;
					}
				}
				else // it is a TreeNode (not VirtualMachineThreeNode) and likely to have children
				{
					foreach (VirtualMachineTreeNode vmNode in topLevelNode.Nodes)
					{
						if (vmNode.Text == SelectedVirtualMachineUserControl.VM.Name)
						{
							tvGroups.SelectedNode = vmNode;
							vmNode.EnsureVisible();
							vmNode.TreeView.Select();
							return;
						}
					}
				}
			}
		}
		#endregion



		private void btnSettings_Click(object sender, EventArgs e)
		{

		}

		private void btnRemove_Click(object sender, EventArgs e)
		{

		}

		private void btnGroups_Click(object sender, EventArgs e)
		{
			if (this.Controls.Contains(tvGroups)) // form is expanded
			{
				HideVirtualMachineGroupsTree();
				btnGroups.Text = "&Groups >>";
			}
			else // add control and reverse Details button arrows
			{
				ShowVirtualMachineGroupsTree();
				btnGroups.Text = "&Groups <<";
			}
		}
		
		#region Groups TreeView Helpers
		
		private int initialFormWidth = 355;
		private int expandedFormWidth = 620;
		private int initialGroupsTreeWidth = 260;

		private void ShowVirtualMachineGroupsTree()
		{
			this.Width = expandedFormWidth;
			this.Controls.Add(tvGroups);
			tvGroups.Width = initialGroupsTreeWidth;
			tvGroups.Height = flpVMs.Height;
			tvGroups.Visible = true;
			tvGroups.ResumeLayout();
		}

		private void HideVirtualMachineGroupsTree()
		{
			tvGroups.Visible = false;
			tvGroups.SuspendLayout();
			this.Controls.Remove(tvGroups);

			this.Width = initialFormWidth;
		}

		// TODO: Make this a bit more robust and offer error messages if result is unexpected
		string[] groupList = System.Configuration.ConfigurationManager.AppSettings["GroupStrings"].Split('|');

		private void InitVirtualMachineGroupsTree()
		{
			foreach (string group in groupList)
			{
				if (!string.IsNullOrEmpty(group))
				{
					tvGroups.Nodes.Add(group);
				}
			}
		}

		private void AddVirtualMachineToGroupsTree(IVirtualMachine vm)
		{
			foreach (TreeNode n in tvGroups.Nodes)
			{
				switch (System.Configuration.ConfigurationManager.AppSettings["GroupType"])
				{
					case "StartsWith":
						if (vm.Name.ToUpper().StartsWith(n.Text.ToUpper()))
						{
							n.Nodes.Add(new VirtualMachineTreeNode(vm));
							return;
						}
						break;
					case "EndsWith":
						if (vm.Name.ToUpper().EndsWith(n.Text.ToUpper()))
						{
							n.Nodes.Add(new VirtualMachineTreeNode(vm));
							return;
						}
						break;
					case "SubString":
						if (vm.Name.ToUpper().Contains(n.Text.ToUpper()))
						{
							n.Nodes.Add(new VirtualMachineTreeNode(vm));
							return;
						}
						break;
				}
			}
			tvGroups.Nodes.Add(new VirtualMachineTreeNode(vm));
		}

		private class VirtualMachineTreeNode : TreeNode
		{
			public readonly IVirtualMachine VM;
			public VirtualMachineTreeNode(IVirtualMachine vm)
			{
				this.VM = vm;
				this.Text = vm.Name;
			}
		}

		#endregion

		private void tvGroups_AfterSelect(object sender, TreeViewEventArgs e)
		{
			VirtualMachineTreeNode vmNode = e.Node as VirtualMachineTreeNode;
			
			// if the selected node is null, it does not refer to a VM, but rather a group (aka container)
			mnuCreateDiffVM.Enabled = (vmNode != null);

			// Show state of currently selected VM (even if non-vm node is selected in tree)
			if (vmNode != null)
			{
				ucActions.VirtualMachine = vmNode.VM;
			}
			else
			{
				if (SelectedVirtualMachineUserControl != null)
				{
					ucActions.VirtualMachine = SelectedVirtualMachineUserControl.VM;
				}
			}

			// Synchronize the selections in the TreeView and FlowLayoutPanel
			if (SelectedVirtualMachineUserControl != null &&
				vmNode != null && // some nodes in the tree are not VMs
				vmNode.VM == SelectedVirtualMachineUserControl.VM)
			{
				// already in sync, do nothing.
				return;
			}

			foreach (VirtualMachineUserControl ucVM in flpVMs.Controls)
			{
				if (ucVM.VM.Name == e.Node.Text)
				{
					ucVM.OnSelect();
					flpVMs.ScrollControlIntoView(ucVM);
					break;
				}
			}
		}

		private void mnuCreateDiffVM_Click(object sender, EventArgs e)
		{
			FolderBrowserDialog dlgFolder = new FolderBrowserDialog();
			dlgFolder.Description = "Select folder for Difference drives.";
			dlgFolder.ShowNewFolderButton = true;
			DialogResult response = dlgFolder.ShowDialog();
			if (response == DialogResult.OK)
			{
				string diffFolder = dlgFolder.SelectedPath;

				dlgFolder.Description = "Select folder for VMC/VSV files.";
				response = dlgFolder.ShowDialog();
				if (response == DialogResult.OK)
				{
					string vmcFolder = dlgFolder.SelectedPath;

					VirtualMachineTreeNode tn = tvGroups.SelectedNode as VirtualMachineTreeNode;
					IVirtualMachine targetVM = tn.VM;

					CreateDiffVM(diffFolder, vmcFolder, targetVM, Guid.NewGuid().ToString());
				}
			}
		}

		private static void CreateDiffVM(string diffFolder, string vmcFolder, IVirtualMachine targetVM, string uniqueAdditionToName)
		{
			IVirtualMachine newVM = targetVM.VS.CreateVirtualMachine(targetVM.Name + " " + uniqueAdditionToName, vmcFolder);
			newVM.Memory = targetVM.Memory;
			newVM.Undoable = false;
			foreach (IHardDiskConnection vhdConn in targetVM.HardDiskConnections)
			{
				FileInfo f = new FileInfo(vhdConn.VirtualHardDisk.File);
				string diffFileName = diffFolder + "\\" + f.Name.Replace(".vhd", "."+uniqueAdditionToName+".diff.vhd");
				ITask t = targetVM.VS.CreateDifferencingVirtualHardDisk(diffFileName, f.FullName);
				t.WaitForCompletion(-1);
				newVM.AddHardDiskConnection(diffFileName, vhdConn.BusType, vhdConn.BusNumber, vhdConn.DeviceNumber);
			}
		}

		private void mnuOptions_Click(object sender, EventArgs e)
		{
			VirtualServerOptionsForm frmVSO = new VirtualServerOptionsForm();
			frmVSO.Show();
		}

		private void mnuAddExisting_Click(object sender, EventArgs e)
		{
			dlgOpenFile.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
			dlgOpenFile.CheckFileExists = true;
			dlgOpenFile.Multiselect = true;
			DialogResult userResponse = dlgOpenFile.ShowDialog();
			if (userResponse == DialogResult.OK)
			{
				foreach (string targetVMC in dlgOpenFile.FileNames)
				{
					this.vsm.VirtualServer.AddExistingVMC(targetVMC);
				}
			}
		}

		private void mnuMigrateVpcVMs_Click(object sender, EventArgs e)
		{
			string[] targets = Delsink.VirtualPC.GetVirtualPcRegisteredVirtualMachineList();
			foreach (string vmc in targets)
			{
				DialogResult response = MessageBox.Show("Add this VM?" + Environment.NewLine + vmc, "VPC Migration", MessageBoxButtons.YesNo, MessageBoxIcon.Question);
				if (response == DialogResult.Yes)
				{
					this.vsm.VirtualServer.AddExistingVMC(vmc);
				}
			}
		}

		private void mnuCreateMultipleDiffVMs_Click(object sender, EventArgs e)
		{
			dlgFolder.Description = "Select folder for Difference drives.";
			dlgFolder.ShowNewFolderButton = true;
			DialogResult response = dlgFolder.ShowDialog();
			if (response == DialogResult.OK)
			{
				string diffFolder = dlgFolder.SelectedPath;

				dlgFolder.Description = "Select folder for VMC/VSV files.";
				response = dlgFolder.ShowDialog();
				if (response == DialogResult.OK)
				{
					string vmcFolder = dlgFolder.SelectedPath;

					VirtualMachineTreeNode tn = tvGroups.SelectedNode as VirtualMachineTreeNode;
					IVirtualMachine targetVM = tn.VM;

					for (int ix = 0; ix < 20; ix++)
					{
						CreateDiffVM(diffFolder, vmcFolder, targetVM, "Lab " + (ix+1).ToString("00"));
					}
				}
			}
		}

		private void mnuRefresh_Click(object sender, EventArgs e)
		{
			foreach (VirtualMachineUserControl vmuc in flpVMs.Controls)
			{
				vmuc.UpdateUI(true);
			}
		}

		private void vsm_TaskCompleted(object sender, TaskCompletedEventArgs e)
		{
			IVirtualMachine vm = e.Task.VM;

			foreach (VirtualMachineUserControl ucVM in this.flpVMs.Controls)
			{
				if (ucVM.VM == vm)
				{
					ucVM.UpdateUI();
				}
			}
			this.ucActions.UpdateUI();
		}

		private void vsm_TaskProgressChanged(object sender, TaskProgressChangedEventArgs e)
		{
			IVirtualMachine vm = e.Task.VM;

			foreach (VirtualMachineUserControl ucVM in this.flpVMs.Controls)
			{
				if (ucVM.VM == vm)
				{
					ucVM.UpdateUI();
				}
			}
			this.ucActions.UpdateUI();
		}

		private void mnuShowPendingTasks_Click(object sender, EventArgs e)
		{
			PendingTasksForm frmPendingTasks = new PendingTasksForm(this.vsm);
			frmPendingTasks.Show();
		}
	}
}